program grmn2cha;

{
************************************************************

GRMN2CHA - File Format Conversion CHART <-> GARMIN (PCX5 & G7TO)


(c) Heiko Purnhagen 1996-1998
e-mail: purnhage@tnt.uni-hannover.de
WWW:    http://www.fet.uni-hannover.de/~purnhage/


All usual disclaimers apply!

This program may be copied according to the GNU General Public Licence.

Please check README for further details.


Changes:

960720   born
970606   GNU GPL version
980508   added TEXT handling -> ignored
         added G7TO format
         wpt unique name / description
980603   command line support
980606   fixed bug
980621   garmin/g7to "square" waypoint icon

************************************************************
}


{ important options: b- }
{ suggested options: n- e- r- d- l- }


uses
  crt,dos;

const
  title = 'grmn2cha - Heiko Purnhagen 21-jun-98';
  maxwpts = 2000;
  modestr = 'c7 cg 7c gc w7 wg 7w gw';

type
  llt = record
          lat,lon : real;
        end;

var
  fi : text;
  fo : text;
  i : integer;
  modechart : boolean;
  modeg2c : boolean;
  modeg7to : boolean;
  outfile : boolean;

  wpts : array[1..maxwpts] of string[6];

  tllformat : integer;
  tlllength : integer;
  tllgrmn : boolean;
  tllg7to : boolean;


{ lat,lon <-> string }

function ang2str (ang : real; h3 : boolean; len : integer) : string;
  var
    h,m,s : real;
    sh,sm,ss : string;
    l,h1,h2,m1,m2,s1,s2 : integer;
begin
  if tllgrmn then
  begin
    len := len+1;
    if tllformat >= 2 then
      len := len+1;
    if tllformat >= 3 then
      len := len+1;
  end;
  h := ang/pi*180;
  sm := '';
  ss := '';
  if h3 then
    l := 3
  else
    l := 2;
  h1 := l;
  h2 := 0;
  if tllformat = 1 then
  begin
    if len >= l+3 then
    begin
      h1 := len-1;
      h2 := len-l-2;
    end;
  end
  else
  begin
    m := frac(h)*60.0;
    h := trunc(h);
  end;
  if tllformat <> 1 then
  begin
    len := len-l-1;
    m1 := 2;
    m2 := 0;
    if tllformat <> 3 then
    begin
      if len >= 5 then
      begin
        m1 := len-1;
        m2 := len-4;
      end;
    end
    else
    begin
      s := frac(m)*60.0;
      m := trunc(m);
    end;
    if tllformat = 3 then
    begin
      len := len-3;
      s1 := 2;
      s2 := 0;
      if len >= 5 then
      begin
        s1 := len-1;
        s2 := len-4;
      end;
      str (s:s1:s2,ss);
      if ss[1] = '6' then
      begin
        str (0.0:s1:s2,ss);
        m := m+1;
      end;
      if ss[1] = ' ' then
        ss[1] := '0';
      if not tllgrmn then
        ss := ss+'"';
    end;
    str (m:m1:m2,sm);
    if sm[1] = '6' then
    begin
      str (0.0:m1:m2,sm);
      h := h+1;
    end;
    if sm[1] = ' ' then
      sm[1] := '0';
    if not tllgrmn then
      sm := sm+'''';
  end;
  str (h:h1:h2,sh);
  if l = 3 then
    if sh[2] = ' ' then
      sh[2] := '0';
  if sh[1] = ' ' then
    sh[1] := '0';
  if not tllgrmn then
    sh := sh+'';
  if tllgrmn and tllg7to then
    ang2str := sh+' '+sm
  else
    ang2str := sh+sm+ss;
end;

function ll2str (ll : llt) : string;
  var
    s,ss : string;
    lt : real;
    len : integer;
begin
  len := tlllength div 2;
  if ll.lat >= 0 then
  begin
    lt := ll.lat;
    ss := 'N';
  end
  else
  begin
    lt := -ll.lat;
    ss := 'S';
  end;
  if not tllgrmn then
    s := ang2str(lt,false,len-2)+ss
  else
    s := ss+ang2str(lt,false,len-2);
  if ll.lon >= 0 then
  begin
    lt := ll.lon;
    ss := 'E';
  end
  else
  begin
    lt := -ll.lon;
    ss := 'W';
  end;
  if not tllgrmn then
    ll2str := s+' '+ang2str(lt,true,len-1)+ss
  else
    ll2str := s+' '+ss+ang2str(lt,true,len-1);
end;

function str2ang (s : string; var ang : real) : boolean;
{ OK -> true, ERROR -> false}
  const
    tab = #9;
  var
    ts : string;
    hh,mm,ss : real;
    p,c : integer;
begin
  str2ang := false;
  hh := 0;
  mm := 0;
  ss := 0;
  while (s <> '') and (s[length(s)] in [' ',tab]) do
    delete (s,length(s),1);
  while (s <> '') and (s[1] in [' ',tab]) do
    delete (s,1,1);
  if s = '' then
    exit;
  p := pos('',s);
  if p = 0 then
    p := pos(' ',s);
  if p = 0 then
    ts := s
  else
    ts := copy(s,1,p-1);
  val (ts,hh,c);
  if c <> 0 then
    exit;
  if hh < 0 then
    exit;
  if p > 0 then
  begin
    delete (s,1,p);
    while (s <> '') and (s[1] in [' ',tab]) do
      delete (s,1,1);
    if s <> '' then
    begin
      p := pos('''',s);
      if p = 0 then
        p := pos(' ',s);
      if p = 0 then
        ts := s
      else
        ts := copy(s,1,p-1);
      val (ts,mm,c);
      if c <> 0 then
        exit;
      if (frac(hh) <> 0) or (mm < 0) or (mm > 60) then
        exit;
      if p > 0 then
      begin
        delete (s,1,p);
        while (s <> '') and (s[1] in [' ',tab]) do
          delete (s,1,1);
        if s <> '' then
        begin
          p := pos('"',s);
          if p = 0 then
            p := pos(' ',s);
          if p = 0 then
            ts := s
          else
            ts := copy(s,1,p-1);
          val (ts,ss,c);
          if c <> 0 then
            exit;
          if (frac(mm) <> 0) or (ss < 0) or (ss > 60) then
            exit;
          if p > 0 then
          begin
            delete (s,1,p);
            while (s <> '') and (s[1] in [' ',tab]) do
              delete (s,1,1);
            if s <> '' then
              exit;
          end;
        end;
      end;
    end;
  end;
  ang := (hh+mm/60+ss/3600)/180*pi;
  str2ang := true;
end;

function str2ll (s : string; var ll : llt) : boolean;
{ OK -> true, ERROR -> false}
  var
    ss : string;
    sign : integer;
    p1,p2 : integer;
begin
  str2ll := false;
  p1 := pos('N',s);
  if p1 = 0 then
    p1 := pos('n',s);
  sign := 1;
  if p1 = 0 then
  begin
    p1 := pos('S',s);
    if p1 = 0 then
      p1 := pos('s',s);
    sign := -1;
  end;
  if p1 = 0 then
    exit;
  if not str2ang(copy(s,1,p1-1),ll.lat) then
      exit;
  ll.lat := sign*ll.lat;
  if (ll.lat > pi/2) or (ll.lat < -pi/2) then
    exit;
  p2 := pos('E',s);
  if p2 = 0 then
    p2 := pos('e',s);
  sign := 1;
  if p2 = 0 then
  begin
    p2 := pos('W',s);
    if p2 = 0 then
      p2 := pos('w',s);
    sign := -1;
  end;
  if p2 = 0 then
    exit;
  if not str2ang(copy(s,p1+1,p2-p1-1),ll.lon) then
      exit;
  ll.lon := sign*ll.lon;
  if (ll.lon > 2*pi) or (ll.lon < -2*pi) then   { avoids problems at 180 }
    exit;
  str2ll := true;
end;


{ garmin }

function readgrmnll (s : string; var ll : llt) : boolean;
{ OK->true error->false }
  var
    hh,mm : real;
    m7 : integer;
    c : integer;
begin
  { pcx5: 'N5927.0000 E01027.0000' }
  { g7to: 'N59 27.0000 E010 27.0000' }
  m7 := 0;
  if modeg7to then
    m7 := 1;
    readgrmnll := false;
  if length(s) <> 22+2*m7 then
    exit;
  val(copy(s,2,2),hh,c);
  if c <> 0 then
    exit;
  val(copy(s,4+m7,7),mm,c);
  if c <> 0 then
    exit;
  if (s[1] <> 'N') and (s[1] <> 'S') then
    exit;
  ll.lat := (hh+mm/60)/180*pi;
  if s[1] = 'S' then
    ll.lat := -ll.lat;
  val(copy(s,13+m7,3),hh,c);
  if c <> 0 then
    exit;
  val(copy(s,16+2*m7,7),mm,c);
  if c <> 0 then
    exit;
  if (s[12+m7] <> 'E') and (s[12+m7] <> 'W') then
    exit;
  ll.lon := (hh+mm/60)/180*pi;
  if s[12+m7] = 'W' then
    ll.lon := -ll.lon;
  readgrmnll := true;
end;

function writegrmnll (ll : llt) : string;
begin
  tllformat := 2;
  tlllength := 22;
  tllgrmn := true;
  tllg7to := modeg7to;
  writegrmnll := ll2str(ll);
  { pcx5: 'N5927.0000 E01027.0000' }
  { g7to: 'N59 27.0000 E010 27.0000' }
end;

function grmnwptchar (c : char; space : boolean) : char;
begin
  c := upcase(c);
  if not (((c >= '0') and (c <= '9')) or
          ((c >= 'A') and (c <= 'Z')) or
          (c = '-') or (space and (c = ' '))) then
    c := '-';
  grmnwptchar := c;
end;

procedure grmnwptname (var name,descr : string);
  var
    i : integer;
begin
  name := copy(name+'      ',1,6);
  for i := 1 to length(name) do
    name[i] := grmnwptchar(name[i],false);
  if not modeg7to then
  begin
    descr := copy(descr+'                                        ',1,40);
    for i := 1 to length(descr) do
      descr[i] := grmnwptchar(descr[i],true);
  end;
end;

procedure writegrmnfilehdr;
begin
  if not modeg7to then
  begin
    writeln (fo,'H  SOFTWARE NAME & VERSION');
    writeln (fo,'I  PCX5 2.01');
    writeln (fo);
    writeln (fo,'H  R DATUM                IDX DA            DF            DX            DY            DZ');
    writeln (fo,'M  E WGS 84               100 +0.000000e+00 +0.000000e+00 +0.000000e+00 +0.000000e+00 +0.000000e+00');
    writeln (fo);
    writeln (fo,'H  COORDINATE SYSTEM');
    writeln (fo,'U  LAT LON DM');
    writeln (fo);
  end;
end;

procedure writegrmntrkhdr (name : string);
begin
  if not modeg7to then
    writeln (fo,'H  LATITUDE   LONGITUDE   DATE      TIME     ALT    ;',name)
  else
    writeln (fo,'N  New Track Start');
end;

procedure writegrmntrk (ll : llt);
begin
  write (fo,'T  ');
  write (fo,writegrmnll(ll),' ');
  if not modeg7to then
    writeln (fo,'01-JAN-80 00:00:00 -9999')
  else
    writeln (fo,'Tue Jan 01 00:00:00 1980');
end;

procedure writegrmnwpthdr;
begin
  if not modeg7to then
  begin
    write (fo,'H  IDNT   LATITUDE   LONGITUDE   DATE      TIME     ALT   DESCRIPTION                              PROXIMITY ');
    writeln (fo,';waypts');
  end;
end;

procedure writegrmnwpt (name,descr : string; ll : llt);
begin
  { first use grmnwptname(name,descr) }
  write (fo,'W  ',copy(name,1,6):6,' ');
  write (fo,writegrmnll(ll),' ');
  if not modeg7to then
  begin
    write (fo,'01-JAN-80 00:00:00 -9999 ');
    writeln (fo,copy(descr,1,40):40,' 0.00000e+00');
  end
  else
  begin
    write (fo,'Tue Jan 01 00:00:00 1980 ');
    writeln (fo,descr,';46;3');
  end;
end;


{ main procedures }

procedure chartg2c;
  var
    ll : llt;
    srcline : integer;
    numline,numll : integer;
    itemflag : boolean;
    s : string;
    m7 : integer;
begin
  writeln ('converting GARMIN track -> CHART chart/log ...');
  tllformat := 2;
  tlllength := 26;
  tllgrmn := false;
  m7 := 0;
  if modeg7to then
    m7 := 1;
  if outfile then
    writeln (fo,'. grmn2cha chart/log');
  srcline := 0;
  numline := 0;
  numll := 0;
  itemflag := true;
  while not eof (fi) do
  begin
    readln (fi,s);
    inc (srcline);
    if s[1] = 'T' then
    begin
      if not readgrmnll(copy(s,4,22+2*m7),ll) then
      begin
        writeln ('error: lat lon format (line ',srcline,')');
        exit;
      end;
      if itemflag then
      begin
        if outfile then
        begin
          writeln (fo,'line');
          writeln (fo,'garmin track line ',copy(s,27+2*m7,18+6*m7));
        end;
        inc (numline);
      end;
      itemflag := false;
      s := ll2str(ll);
      if outfile then
        writeln (fo,s);
      inc (numll);
    end
    else
      itemflag := true;
  end;
  writeln ('input file:   ',srcline:4,' lines');
  writeln ('chart lines:  ',numline:4);
  writeln ('total points: ',numll:4);
end;

procedure chartc2g;
  var
    ll,ll1 : llt;
    s,ss,st : string;
    polyflag,firstflag,itemflag : boolean;
    srcline : integer;
    numline,numpoly,nummark,numtext,numll : integer;
begin
  writeln ('converting CHART chart/log -> GARMIN track ...');
  if outfile then
    writegrmnfilehdr;
  polyflag := false;
  firstflag := true;
  itemflag := true;
  numline := 0;
  numpoly := 0;
  nummark := 0;
  numtext := 0;
  numll := 0;
  srcline := 0;
  while not eof(fi) do
  begin
    readln (fi,s);
    inc (srcline);
    if (s <> '') and (s[1] <> '.') then
    begin
      st := copy(s,1,4);
      if (st = 'line') or (st = 'poly') or (st = 'mark') then
      begin
        if polyflag then
        begin
          if outfile then
            writegrmntrk (ll1);
          inc (numll);
        end;
        if not firstflag and outfile then
          writeln (fo);
        firstflag := false;
        if st = 'line' then
          inc (numline)
        else if st = 'poly' then
          inc (numpoly)
        else
          inc (nummark);
        readln (fi,ss);
        inc (srcline);
        if outfile then
          writegrmntrkhdr (s+' '+ss);
        repeat
          readln (fi,ss);
          inc (srcline);
        until (ss <> '') and (ss[1] <> '.');
        if not str2ll(ss,ll) then
        begin
          writeln ('error: lat lon format (line ',srcline,')');
          exit;
        end;
        if outfile then
          writegrmntrk (ll);
        inc (numll);
        polyflag := (st = 'poly');
        itemflag := (st = 'mark');
        ll1 := ll;
      end
      else if st = 'text' then
      begin
        readln (fi,ss);
        inc (srcline);
        repeat
          readln (fi,ss);
          inc (srcline);
        until (ss <> '') and (ss[1] <> '.');
        inc (numtext);
      end
      else
      begin
        if itemflag then
        begin
          writeln ('error: expecting line/poly/mark/text (line ',srcline,')');
          exit;
        end;
        if not str2ll(s,ll) then
        begin
          writeln ('error: lat lon format (line ',srcline,')');
          exit;
        end;
        if outfile then
          writegrmntrk (ll);
        inc (numll);
      end;
    end;
  end;
  if polyflag then
  begin
    if outfile then
      writegrmntrk (ll1);
    inc (numll);
  end;
  writeln ('input file:    ',srcline:4,' lines');
  writeln ('chart lines:   ',numline:4);
  writeln ('chart polys:   ',numpoly:4);
  writeln ('chart marks:   ',nummark:4);
  writeln ('chart texts:   ',numtext:4,' (ignored)');
  writeln ('chart points:  ',numll-numpoly+numtext:4);
  writeln ('garmin points: ',numll:4);
end;

procedure wptg2c;
  var
    ll : llt;
    srcline : integer;
    numwpts : integer;
    s,name : string;
    m7 : integer;
begin
  writeln ('converting GARMIN waypoints -> CHART waypoints/memo ...');
  tllformat := 2;
  tlllength := 26;
  tllgrmn := false;
  m7 := 0;
  if modeg7to then
    m7 := 1;
  if outfile then
    writeln (fo,'. grmn2cha waypoints');
  srcline := 0;
  numwpts := 0;
  while not eof (fi) do
  begin
    readln (fi,s);
    inc (srcline);
    if s[1] = 'W' then
    begin
      name := copy(s,4,6);
      if not readgrmnll(copy(s,11,22+2*m7),ll) then
      begin
        writeln ('error: lat lon format (line ',srcline,')');
        exit;
      end;
      name := name+' '+copy(s,59+2*m7,40);
      s := ll2str(ll);
      if outfile then
      begin
        writeln (fo,'mark');
        writeln (fo,name);
        writeln (fo,s);
      end;
      inc (numwpts);
    end;
  end;
  writeln ('input file: ',srcline:4,' lines');
  writeln ('waypoints:  ',numwpts:4);
end;

procedure wptc2g;
  var
    ll : llt;
    s,ss,n : string;
    srcline : integer;
    numwpts : integer;
    name,descr : string;
    i : integer;
    unique,failed : boolean;
    c : char;
begin
  writeln ('converting CHART waypoints/memo -> GARMIN waypoints ...');
  if outfile then
  begin
    writegrmnfilehdr;
    writegrmnwpthdr;
  end;
  numwpts := 0;
  srcline := 0;
  while not eof(fi) do
  begin
    readln (fi,s);
    inc (srcline);
    if (s <> '') and (s[1] <> '.') then
    begin
      if copy(s,1,4) <> 'mark' then
      begin
        writeln ('error: expecting mark (line ',srcline,')');
        exit;
      end
      else
      begin
        inc (numwpts);
        if numwpts > maxwpts then
        begin
          writeln ('error: more than ',maxwpts,' waypoints');
          exit;
        end;
        readln (fi,s);
        inc (srcline);
        repeat
          readln (fi,ss);
          inc (srcline);
        until (ss <> '') and (ss[1] <> '.');
        if not str2ll(ss,ll) then
        begin
          writeln ('error: lat lon format (line ',srcline,')');
          exit;
        end;
        name := s;
        if copy(s,7,1) = ' ' then
          descr := copy(s,8,40)
        else
          descr := s;
        c := ' ';
        grmnwptname (name,descr);
        n := name;
        failed := false;
        repeat
          unique := true;
          for i := 1 to numwpts-1 do
            if name = wpts[i] then
              unique := false;
          if not unique then
          begin
            if c = ' ' then
              c := '0'
            else if c = '9' then
              c := 'A'
            else if c = 'Z' then
              failed := true
            else
              inc (c);
            name[6] := c;
            grmnwptname (name,descr);
          end;
        until unique or failed;
        if c <> ' ' then
          writeln ('note: "',n,'" not unique, renaming to "',name,'"');
        if failed then
          writeln ('warning: "',name,'" still not unique');
        wpts[numwpts] := name;
        if outfile then
          writegrmnwpt (name,descr,ll);
      end;
    end;
  end;
  writeln ('input file: ',srcline:4,' lines');
  writeln ('waypoints:  ',numwpts:4);
end;


{ main }

begin
  writeln (title);
  if (paramcount<>2) and (paramcount<>3) then
  begin
    writeln;
    writeln ('Convert between CHART chart/waypts/memo files and');
    writeln ('GARMIN track/waypts file (PCX5 & G7TO format).');
    writeln ('Show input file statistics. Output file optional.');
    writeln;
    writeln ('grmn2cha <mode> <input file> [<output file>]');
    writeln;
    writeln ('modes: c7  chart/log   -> G7TO track');
    writeln ('       cg  chart/log   -> PCX5 track');
    writeln ('       7c  G7TO track  -> chart');
    writeln ('       gc  PCX5 track  -> chart');
    writeln ('       w7  waypts/memo -> G7TO waypts');
    writeln ('       wg  waypts/memo -> PCX5 waypts');
    writeln ('       7w  G7TO waypts -> waypts/memo');
    writeln ('       gw  PCX5 waypts -> waypts/memo');
    exit;
  end;

  i := 0;
  repeat

    if paramstr(1) = copy(modestr,3*i+1,2) then
    begin
      modechart := (i and 4 = 0);
      modeg2c := (i and 2 = 2);
      modeg7to := (i and 1 = 0);
      i := 9;
    end
    else
      inc(i);
  until i > 7;
  if i = 8 then
  begin
    writeln ('unknown mode ',paramstr(1));
    exit;
  end;

  assign (fi,paramstr(2));
  {$i-}
  reset (fi);
  {$i+}
  if ioresult <> 0 then
  begin
    writeln ('error: can''t open input file');
    exit;
  end;

  outfile := paramcount = 3;
  if outfile then
  begin
    assign (fo,paramstr(3));
    {$i-}
    rewrite (fo);
    {$i+}
    if ioresult <> 0 then
    begin
      writeln ('error: can''t open output file');
      exit;
    end;
  end;

  if modechart then
  begin
    if modeg2c then
      chartg2c
    else
      chartc2g;
  end
  else
  begin
    if modeg2c then
      wptg2c
    else
      wptc2g;
  end;

  {$i-}
  close (fi);
  {$i+}
  if ioresult <> 0 then
  begin
    writeln ('error: can''t close input file');
    exit;
  end;
  if outfile then
  begin
    {$i-}
    close (fo);
    {$i+}
    if ioresult <> 0 then
    begin
      writeln ('error: can''t close output file');
      exit;
    end;
  end;
end.

